/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2022, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/** \file inte_adapt_cern.h
    \brief File defining \ref o2scl::inte_adapt_cern
*/
#ifndef O2SCL_CERN_ADAPT_H
#define O2SCL_CERN_ADAPT_H

#include <o2scl/misc.h>
#include <o2scl/funct_multip.h>
#include <o2scl/inte.h>
#include <o2scl/inte_gauss56_cern.h>
#include <o2scl/string_conv.h>
#include <o2scl/vector_derint.h>
 
namespace o2scl {

  /** \brief Adaptive integration (CERNLIB)
    
      Uses a base integration object (default is \ref
      inte_gauss56_cern) to perform adaptive integration by
      automatically subdividing the integration interval. At each
      step, the interval with the largest absolute uncertainty is
      divided in half. The routine succeeds if the absolute tolerance
      is less than \ref tol_abs or if the relative tolerance is less
      than \ref tol_rel, i.e.
      \f[
      \mathrm{err}\leq\mathrm{tol\_abs}~\mathrm{or}~
      \mathrm{err}\leq\mathrm{tol\_rel}\cdot|I|
      \f]
      where \f$ I \f$ is the current estimate for the integral and \f$
      \mathrm{err} \f$ is the current estimate for the uncertainty. If
      the number of subdivisions exceeds the template parameter \c
      nsub, the error handler is called, since the integration may not
      have been successful. The number of subdivisions used in the
      last integration can be obtained from get_nsubdivisions().

      The template parameter \c nsub, is the maximum number of
      subdivisions. It is automatically set to 100 in the original
      CERNLIB routine, and defaults to 100 here. The default base
      integration object is of type \ref inte_gauss56_cern. This is the
      CERNLIB default, but can be modified by calling set_inte().

      This class is based on the CERNLIB routines RADAPT and
      DADAPT which are documented at
      http://wwwasdoc.web.cern.ch/wwwasdoc/shortwrupsdir/d102/top.html
      
      \future 
      - Allow user to set the initial subdivisions?
      - It might be interesting to directly compare the performance
      of this class to \ref o2scl::inte_qag_gsl .
      - There is a fixme entry in the code which could be resolved.
      - Output the point where most subdividing was required?
  */
  template<class func_t=funct,
           class def_inte_t=inte_gauss56_cern<funct,double,
                                              inte_gauss56_coeffs_double>,
           size_t nsub=100, class fp_t=double>
  class inte_adapt_cern : public inte<func_t,fp_t> {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// Lower end of subdivision
    fp_t xlo[nsub];

    /// High end of subdivision
    fp_t xhi[nsub];

    /// Value of integral for subdivision
    fp_t tval[nsub];

    /// Squared error for subdivision
    fp_t ters[nsub];
      
    /// Previous number of subdivisions
    int prev_subdiv;

    /// The base integration object
    inte<func_t,fp_t> *it;
      
#endif
      
  public:
  
    inte_adapt_cern() {
      nsubdiv=1;
      prev_subdiv=0;

      it=&def_inte;
    }

    /// \name Basic usage
    //@{
    /** \brief Integrate function \c func from \c a to \c b
        giving result \c res and error \c err
    */
    virtual int integ_err(func_t &func, fp_t a, fp_t b,
                          fp_t &res, fp_t &err) {
        
      fp_t tvals=0.0, terss, xlob, xhib, yhib=0.0, te, root=0.0;
      int i, nsubdivd;
      fp_t two=2.0;

      if (nsubdiv==0) {
        if (prev_subdiv==0) {
          // If the previous binning was requested, but
          // there is no previous binning stored, then
          // just shift to automatic binning
          nsubdivd=1;
        } else {
          tvals=0.0;
          terss=0.0;
          for(i=0;i<prev_subdiv;i++) {
            it->integ_err(func,xlo[i],xhi[i],tval[i],te);
            ters[i]=te*te;
            tvals+=tval[i];
            terss+=ters[i];
          }
          err=sqrt(two*terss);
          res=tvals;
          return 0;
        }
      }
  
      // Ensure we're not asking for too many divisions
      if (nsub<nsubdiv) {
        nsubdivd=nsub;
      } else {
        nsubdivd=nsubdiv;
      }

      // Compute the initial set of intervals and integral values
      xhib=a;
      fp_t bin=(b-a)/((fp_t)nsubdivd);
      for(i=0;i<nsubdivd;i++) {
        xlo[i]=xhib;
        xlob=xlo[i];
        xhi[i]=xhib+bin;
        if (i==nsubdivd-1) xhi[i]=b;
        xhib=xhi[i];
        it->integ_err(func,xlob,xhib,tval[i],te);
        ters[i]=te*te;
      }
      prev_subdiv=nsubdivd;

      for(size_t iter=1;iter<=nsub;iter++) {

        // Compute the total value of the integrand
        // and the squared uncertainty
        tvals=tval[0];
        terss=ters[0];
        for(i=1;i<prev_subdiv;i++) {
          tvals+=tval[i];
          terss+=ters[i];
        }
          
        // Output iteration information
        if (this->verbose>0) {
          std::cout << "inte_adapt_cern Iter: " << iter;
          std::cout.setf(std::ios::showpos);
          std::cout << " Res: " << tvals;
          std::cout.unsetf(std::ios::showpos);
          std::cout << " Err: " << sqrt(two*terss);
          if (this->tol_abs>this->tol_rel*abs(tvals)) {
            std::cout << " Tol: " << this->tol_abs << std::endl;
          } else {
            std::cout << " Tol: " << this->tol_rel*abs(tvals)
                      << std::endl;
          }
          if (this->verbose>1) {
            char ch;
            std::cout << "Press a key and type enter to continue. " ;
            std::cin >> ch;
          }
        }

        // See if we're finished
        root=sqrt(two*terss);
        if (root<=this->tol_abs || root<=this->tol_rel*abs(tvals)) {
          res=tvals;
          err=root;
          this->last_iter=iter;
          return 0;
        }

        // Test if we've run out of intervals
        if (prev_subdiv==nsub) {
          res=tvals;
          err=root;
          this->last_iter=iter;
          std::string s="Reached maximum number ("+itos(nsub)+
            ") of subdivisions in inte_adapt_cern::integ_err().";
          O2SCL_CONV_RET(s.c_str(),exc_etable,this->err_nonconv);
        }

        // Find the subdivision with the largest error
        fp_t bige=ters[0];
        int ibig=0;
        for(i=1;i<prev_subdiv;i++) {
          if (ters[i]>bige) {
            bige=ters[i];
            ibig=i;
          }
        }

        // Subdivide that subdivision further
        xhi[prev_subdiv]=xhi[ibig];
        fp_t xnew=(xlo[ibig]+xhi[ibig])/two;
        xhi[ibig]=xnew;
        xlo[prev_subdiv]=xnew;
        it->integ_err(func,xlo[ibig],xhi[ibig],tval[ibig],te);
        ters[ibig]=te*te;
        it->integ_err(func,xlo[prev_subdiv],
                      xhi[prev_subdiv],tval[prev_subdiv],te);
        ters[prev_subdiv]=te*te;
        prev_subdiv++;

      }

      // FIXME: Should we set an error here, or does this
      // only happen if we happen to need exactly nsub
      // intervals?
      res=tvals;
      err=root;
      return 0;
    }
    //@}

    /// \name Integration object
    //@{
    /// Set the base integration object to use
    int set_inte(inte<func_t,fp_t> &i) {
      it=&i;
      return 0;
    }
      
    /// Default integration object
    def_inte_t def_inte;
    //@}
      
    /// \name Subdivisions
    //@{
    /** \brief Number of subdivisions 

        The options are
        - 0: Use previous binning and do not subdivide further \n
        - 1: Automatic - adapt until tolerance is attained (default) \n
        - n: (n>1) split first in n equal subdivisions, then adapt
        until tolerance is obtained.
    */
    size_t nsubdiv;

    /// Return the number of subdivisions used in the last integration
    size_t get_nsubdivisions() {
      return prev_subdiv;
    }

    /// Return the ith subdivision
    int get_ith_subdivision(size_t i, fp_t &xlow, fp_t &xhigh, 
                            fp_t &value, fp_t &errsq) {
      if (i<prev_subdiv) {
        xlow=xlo[i];
        xhigh=xhi[i];
        value=tval[i];
        errsq=ters[i];
      }
      return 0;
    }
      
    /// Return all of the subdivisions
    template<class vec_t> 
    int get_subdivisions(vec_t &xlow, vec_t &xhigh, vec_t &value, 
                         vec_t &errsq) {

      for(int i=0;i<prev_subdiv;i++) {
        xlow[i]=xlo[i];
        xhigh[i]=xhi[i];
        value[i]=tval[i];
        errsq[i]=ters[i];
      }
      return 0;
    }
    //@}
    
  };
  
  typedef boost::multiprecision::number<
    boost::multiprecision::cpp_dec_float<25>> cpp_dec_float_25;
  typedef boost::multiprecision::number<
    boost::multiprecision::cpp_dec_float<35>> cpp_dec_float_35;
  typedef boost::multiprecision::number<
    boost::multiprecision::cpp_dec_float<50>> cpp_dec_float_50;
  typedef boost::multiprecision::number<
    boost::multiprecision::cpp_dec_float<100>> cpp_dec_float_100;

  typedef
  inte_adapt_cern<funct_ld,inte_gauss56_cern
                  <funct_ld,long double,
                   inte_gauss56_coeffs_long_double>,100,
                  long double> inte_adapt_cern_ld;

  typedef
  inte_adapt_cern<funct_cdf25,inte_gauss56_cern
                  <funct_cdf25,cpp_dec_float_25,
                   inte_gauss56_coeffs_float_50<cpp_dec_float_25>>,1000,
                   cpp_dec_float_25> inte_adapt_cern_cdf25;
  
  typedef
  inte_adapt_cern<funct_cdf35,inte_gauss56_cern
                  <funct_cdf35,cpp_dec_float_35,
                   inte_gauss56_coeffs_float_50<cpp_dec_float_35>>,1000,
                   cpp_dec_float_35> inte_adapt_cern_cdf35;
  
  typedef
  inte_adapt_cern<funct_cdf50,inte_gauss56_cern
                  <funct_cdf50,cpp_dec_float_50,
                   inte_gauss56_coeffs_float_50<cpp_dec_float_50>>,1000,
                   cpp_dec_float_50> inte_adapt_cern_cdf50;
  
  typedef std::function<double(const double &)> funct_cr;
  typedef std::function<long double(const long double &)> funct_cr_ld;
  typedef std::function<cpp_dec_float_25(const cpp_dec_float_25 &)>
  funct_cr_cdf25;
  typedef std::function<cpp_dec_float_35(const cpp_dec_float_35 &)>
  funct_cr_cdf35;
  typedef std::function<cpp_dec_float_50(const cpp_dec_float_50 &)>
  funct_cr_cdf50;
  typedef std::function<cpp_dec_float_100(const cpp_dec_float_100 &)>
  funct_cr_cdf100;
  
  typedef
  inte_adapt_cern<funct_cr,inte_gauss56_cern
                  <funct_cr,double,
                   inte_gauss56_coeffs_double>,100,
                  double> inte_adapt_cern_cr;
  
  typedef
  inte_adapt_cern<funct_cr_ld,inte_gauss56_cern
                  <funct_cr_ld,long double,
                   inte_gauss56_coeffs_long_double>,100,
                  long double> inte_adapt_cern_cr_ld;

  typedef
  inte_adapt_cern<funct_cr_cdf25,inte_gauss56_cern
                  <funct_cr_cdf25,cpp_dec_float_25,
                   inte_gauss56_coeffs_float_50<cpp_dec_float_25>>,1000,
                   cpp_dec_float_25> inte_adapt_cern_cr_cdf25;
  
  typedef
  inte_adapt_cern<funct_cr_cdf35,inte_gauss56_cern
                  <funct_cr_cdf35,cpp_dec_float_35,
                   inte_gauss56_coeffs_float_50<cpp_dec_float_35>>,1000,
                   cpp_dec_float_35> inte_adapt_cern_cr_cdf35;
  
  typedef
  inte_adapt_cern<funct_cr_cdf50,inte_gauss56_cern
                  <funct_cr_cdf50,cpp_dec_float_50,
                   inte_gauss56_coeffs_float_50<cpp_dec_float_50>>,1000,
                   cpp_dec_float_50> inte_adapt_cern_cr_cdf50;
  
  class inte_adapt_cern2 {

  public:
  
    inte_adapt_cern2() {
      tol_rel_multip=-1.0;
      verbose=0;
      pow_tol_func=1.33;
      err_nonconv=true;
      tol_rel=1.0e-8;
      tol_abs=0.0;
      nsub=100;
    }

    int last_iter;

    int nsub;
    
    /** \brief The maximum relative uncertainty for multipreicsion
	integrals (default \f$ -1 \f$)
    */
    double tol_rel_multip;

    /** \brief Power for tolerance of function evaluations in
        multiprecision integrations (default 1.33)
    */
    double pow_tol_func;

    /** \brief The maximum relative uncertainty 
	in the value of the integral (default \f$ 10^{-8} \f$)
    */
    double tol_rel;

    /** \brief The maximum absolute uncertainty 
	in the value of the integral (default \f$ 10^{-8} \f$)
    */
    double tol_abs;

    /** \brief Verbosity parameter
     */
    int verbose;

    /** \brief If true, call the error handler if the integration
        does not succeed (default true)
    */
    bool err_nonconv;
    
    typedef boost::multiprecision::number<
      boost::multiprecision::cpp_dec_float<25>> cpp_dec_float_25;
    typedef boost::multiprecision::number<
      boost::multiprecision::cpp_dec_float<35>> cpp_dec_float_35;
    typedef boost::multiprecision::number<
      boost::multiprecision::cpp_dec_float<50>> cpp_dec_float_50;

    /// \name Nonadaptive integration objects
    //@{
    inte_gauss56_cern<funct,double,
                      inte_gauss56_coeffs_double> def_d;
    inte_gauss56_cern
    <funct_ld,long double,
     inte_gauss56_coeffs_long_double> def_ld;
    inte_gauss56_cern
    <funct_cdf25,cpp_dec_float_25,
     inte_gauss56_coeffs_float_50<cpp_dec_float_25>> def_25;
    inte_gauss56_cern
    <funct_cdf35,cpp_dec_float_35,
     inte_gauss56_coeffs_float_50<cpp_dec_float_35>> def_35;
    inte_gauss56_cern
    <funct_cdf50,cpp_dec_float_50,
     inte_gauss56_coeffs_float_50<cpp_dec_float_50>> def_50;
    //@}
    
    /// \name Basic usage
    //@{
    /** \brief Integrate function \c func from \c a to \c b
        giving result \c res and error \c err
    */
    template<class inte_t, typename func_t, class fp_t>
    int integ_err_funct(inte_t &it, func_t &func, fp_t a, fp_t b,
                        fp_t &res, fp_t &err, double integ_tol) {
      
      // Lower end of subdivision
      fp_t xlo[nsub];
      
      // High end of subdivision
      fp_t xhi[nsub];
      
      // Value of integral for subdivision
      fp_t tval[nsub];
      
      // Squared error for subdivision
      fp_t ters[nsub];
      
      fp_t tvals=0.0, terss, xlob, xhib, yhib=0.0, te, root=0.0;
      int i, nsubdivd, prev_subdiv;
      fp_t two=2.0;

      // Automatic binning
      nsubdivd=1;

      // Compute the initial set of intervals and integral values
      xhib=a;
      fp_t bin=(b-a)/((fp_t)nsubdivd);
      for(i=0;i<nsubdivd;i++) {
        xlo[i]=xhib;
        xlob=xlo[i];
        xhi[i]=xhib+bin;
        if (i==nsubdivd-1) xhi[i]=b;
        xhib=xhi[i];
        it.integ_err2(func,xlob,xhib,tval[i],te);
        ters[i]=te*te;
      }
      prev_subdiv=nsubdivd;

      for(size_t iter=1;iter<=nsub;iter++) {

        // Compute the total value of the integrand
        // and the squared uncertainty
        tvals=tval[0];
        terss=ters[0];
        for(i=1;i<prev_subdiv;i++) {
          tvals+=tval[i];
          terss+=ters[i];
        }
          
        // Output iteration information
        if (this->verbose>0) {
          std::cout << "inte_adapt_cern Iter: " << iter;
          std::cout.setf(std::ios::showpos);
          std::cout << " Res: " << tvals;
          std::cout.unsetf(std::ios::showpos);
          std::cout << " Err: " << sqrt(two*terss);
          if (this->tol_abs>integ_tol*abs(tvals)) {
            std::cout << " Tol: " << this->tol_abs << std::endl;
          } else {
            std::cout << " Tol: " << integ_tol*abs(tvals)
                      << std::endl;
          }
          if (this->verbose>1) {
            char ch;
            std::cout << "Press a key and type enter to continue. " ;
            std::cin >> ch;
          }
        }

        // See if we're finished
        root=sqrt(two*terss);
        if (root<=this->tol_abs || root<=integ_tol*abs(tvals)) {
          res=tvals;
          err=root;
          this->last_iter=iter;
          return 0;
        }

        // Test if we've run out of intervals
        if (prev_subdiv==nsub) {
          res=tvals;
          err=root;
          this->last_iter=iter;
          std::string s="Reached maximum number ("+itos(nsub)+
            ") of subdivisions in inte_adapt_cern::integ_err2().";
          O2SCL_CONV_RET(s.c_str(),exc_etable,this->err_nonconv);
        }

        // Find the subdivision with the largest error
        fp_t bige=ters[0];
        int ibig=0;
        for(i=1;i<prev_subdiv;i++) {
          if (ters[i]>bige) {
            bige=ters[i];
            ibig=i;
          }
        }

        // Subdivide that subdivision further
        xhi[prev_subdiv]=xhi[ibig];
        fp_t xnew=(xlo[ibig]+xhi[ibig])/two;
        xhi[ibig]=xnew;
        xlo[prev_subdiv]=xnew;
        it.integ_err2(func,xlo[ibig],xhi[ibig],tval[ibig],te);
        ters[ibig]=te*te;
        it.integ_err2(func,xlo[prev_subdiv],
                     xhi[prev_subdiv],tval[prev_subdiv],te);
        ters[prev_subdiv]=te*te;
        prev_subdiv++;

      }

      // FIXME: Should we set an error here, or does this
      // only happen if we happen to need exactly nsub
      // intervals?
      res=tvals;
      err=root;
      return 0;
    }
    //@}

    template<typename func_t, class fp_t>
    int integ_err(func_t &func, fp_t a, fp_t b, fp_t &res, fp_t &err) {
      
      int ret=integ_err_funct(func,a,b,res,err,this->tol_rel);
      
      if (ret!=0) {
        if (this->verbose>0) {
          std::cout << "Function inte_adapt_cern2::integ_err() failed."
                    << std::endl;
          std::cout << "Values err,tol_rel: "
                    << err << " " << this->tol_rel << " "
                    << std::endl;
        }
        O2SCL_CONV2_RET("Failed to achieve tolerance in ",
                        "inte_adapt_cern2::integ_err().",o2scl::exc_efailed,
                        this->err_nonconv);
      }
      return 0;
    }

    template <class inte_t, typename func_t, class fp_t>
    int integ_err_int(inte_t &it, func_t &&func, fp_t a, fp_t b, 
                      fp_t &res, fp_t &err, 
                      double integ_tol, double func_tol) {
      
      funct_multip fm2;
      fm2.err_nonconv=false;
      fm2.tol_rel=func_tol;

      std::function<fp_t(fp_t)> fx=[fm2,func](fp_t x) mutable -> fp_t
      { return fm2(func,x); };

      integ_err_funct(it,fx,a,b,res,err,integ_tol);

      if (verbose>1) {
        std::cout << "inte_adapt_cern::integ_err() "
                  << "tols(integ,func),err:\n  "
                  << integ_tol << " "
                  << func_tol << " " << err << std::endl;
      }

      if (err/abs(res)>integ_tol) {
        return 1;
      }
      return 0;
    }
    
    /** \brief Integrate function \c func from \c a to \c b using
        multipreicsion, placing the result in \c res and the error in
        \c err
    */
    template <typename func_t, class fp_t>
    int integ_err_multip(func_t &&func, fp_t a, fp_t b, 
                         fp_t &res, fp_t &err, double integ_tol=-1.0) {
      
      if (integ_tol<=0.0) {
        if (tol_rel_multip<=0.0) {
          integ_tol=pow(10.0,-std::numeric_limits<fp_t>::digits10);
        } else {
          integ_tol=tol_rel_multip;
        }
      } 

      if (verbose>0) {
        std::cout << "int_adapt_cern::integ_err(): set "
                  << "tolerance to: " << integ_tol << std::endl;
      }
      
      // Demand that the function evaluations are higher precision
      double func_tol=pow(integ_tol,pow_tol_func);

      int ret;

      if (integ_tol>pow(10.0,-std::numeric_limits<double>::digits10+3)) {
        if (verbose>0) {
          std::cout << "int_adapt_cern::integ_err(): "
                    << integ_tol << " > "
                    << pow(10.0,-std::numeric_limits<double>::digits10+3)
                    << "\n  for double integration." << std::endl;
        }
        double a_d=static_cast<double>(a);
        double b_d=static_cast<double>(b);
        double res_d, err_d;
        
        ret=integ_err_int(def_d,func,a_d,b_d,res_d,err_d,integ_tol,func_tol);

        std::cout << "H: " << ret << " " << err_d/abs(res_d) << " "
                  << res_d << " " << err_d << " " << integ_tol << std::endl;
        if (ret==0 && err_d/abs(res_d)<integ_tol) {
          res=static_cast<fp_t>(res_d);
          err=static_cast<fp_t>(err_d);
          return 0;
        }
      }

      if (integ_tol>pow(10.0,
                        -std::numeric_limits<long double>::digits10+3)) {
        if (verbose>0) {
          std::cout << "int_adapt_cern::integ_err(): "
                    << integ_tol << " > "
                    << pow(10.0,
                           -std::numeric_limits<long double>::digits10+3)
                    << "\n  for long double integration." << std::endl;
        }
        long double a_ld=static_cast<long double>(a);
        long double b_ld=static_cast<long double>(b);
        long double res_ld, err_ld;
        
        ret=integ_err_int(def_ld,func,a_ld,b_ld,res_ld,err_ld,
                          integ_tol,func_tol);
        
        if (ret==0 && err_ld/abs(res_ld)<integ_tol) {
          res=static_cast<fp_t>(res_ld);
          err=static_cast<fp_t>(err_ld);
          return 0;
        }
      }

      if (integ_tol>pow(10.0,-std::numeric_limits
                        <cpp_dec_float_25>::digits10+3)) {
        if (verbose>0) {
          std::cout << "int_adapt_cern::integ_err(): "
                    << integ_tol << " > "
                    << pow(10.0,-std::numeric_limits
                           <cpp_dec_float_25>::digits10+3)
                    << "\n  for cpp_dec_float_25 integration." << std::endl;
        }
        cpp_dec_float_25 a_cdf25=static_cast<cpp_dec_float_25>(a);
        cpp_dec_float_25 b_cdf25=static_cast<cpp_dec_float_25>(b);
        cpp_dec_float_25 res_cdf25, err_cdf25;
        
        ret=integ_err_int(def_25,func,a_cdf25,b_cdf25,res_cdf25,
                          err_cdf25,integ_tol,func_tol);
        
        if (ret==0 && err_cdf25/abs(res_cdf25)<integ_tol) {
          res=static_cast<fp_t>(res_cdf25);
          err=static_cast<fp_t>(err_cdf25);
          return 0;
        }
      }

      if (integ_tol>pow(10.0,-std::numeric_limits
                        <cpp_dec_float_35>::digits10+3)) {
        if (verbose>0) {
          std::cout << "int_adapt_cern::integ_err(): "
                    << integ_tol << " > "
                    << pow(10.0,-std::numeric_limits
                           <cpp_dec_float_35>::digits10+3)
                    << "\n  for cpp_dec_float_35 integration." << std::endl;
        }
        cpp_dec_float_35 a_cdf35=static_cast<cpp_dec_float_35>(a);
        cpp_dec_float_35 b_cdf35=static_cast<cpp_dec_float_35>(b);
        cpp_dec_float_35 res_cdf35, err_cdf35;
        
        ret=integ_err_int(def_35,func,a_cdf35,b_cdf35,res_cdf35,
                          err_cdf35,integ_tol,func_tol);
        
        if (ret==0 && err_cdf35/abs(res_cdf35)<integ_tol) {
          res=static_cast<fp_t>(res_cdf35);
          err=static_cast<fp_t>(err_cdf35);
          return 0;
        }
      }

      if (integ_tol>pow(10.0,-std::numeric_limits
                        <cpp_dec_float_50>::digits10+3)) {
        if (verbose>0) {
          std::cout << "int_adapt_cern::integ_err(): "
                    << integ_tol << " > "
                    << pow(10.0,-std::numeric_limits
                           <cpp_dec_float_50>::digits10+3)
                    << "\n  for cpp_dec_float_50 integration." << std::endl;
        }
        cpp_dec_float_50 a_cdf50=static_cast<cpp_dec_float_50>(a);
        cpp_dec_float_50 b_cdf50=static_cast<cpp_dec_float_50>(b);
        cpp_dec_float_50 res_cdf50, err_cdf50;
        
        ret=integ_err_int(def_50,func,a_cdf50,b_cdf50,res_cdf50,
                          err_cdf50,integ_tol,func_tol);
        
        if (ret==0 && err_cdf50/abs(res_cdf50)<integ_tol) {
          res=static_cast<fp_t>(res_cdf50);
          err=static_cast<fp_t>(err_cdf50);
          return 0;
        }
      }

      if (verbose>0) {
        std::cout << "inte_adapt_cern::integ_err() "
                  << "failed after cpp_dec_float_50:\n  "
                  << integ_tol << std::endl;
      }
    
      O2SCL_ERR2("Failed to compute with requested accuracy ",
                 "in inte_adapt_cern::integ_err().",
                 o2scl::exc_efailed);
      return o2scl::exc_efailed;
    }

  };

  
}

#endif
